@@ -8,13 +8,9 @@ to sign a wrong transactions.
 Makes sense to run gc.collect() after processing of each scope to free memory.
 """
 from .psbt import *
+from .transaction import hash_amounts, hash_script_pubkeys
 import hashlib
 
-def skip_string(stream) -> bytes:
-    l = compact.read_from(stream)
-    stream.seek(l, 1)
-    return len(compact.to_bytes(l)) + l
-
 def read_write(sin, sout, l=None, chunk_size=32) -> int:
     """Reads l or all bytes from sin and writes to sout"""
     # number of bytes written
@@ -42,9 +38,10 @@ class GlobalTransactionView:
     """
     Global transaction in PSBT is
     - unsigned (with empty scriptsigs)
-    - doesn't have witness 
+    - doesn't have witness
     """
     LEN_VIN = 32 + 4 + 1 + 4 # txid, vout, scriptsig, sequence
+    NUM_VIN_OFFSET = 4 # version
     def __init__(self, stream, offset):
         self.stream = stream
         self.offset = offset
@@ -65,7 +62,7 @@ class GlobalTransactionView:
     @property
     def num_vin(self):
         if self._num_vin is None:
-            self.stream.seek(self.offset + 4)
+            self.stream.seek(self.offset + self.NUM_VIN_OFFSET)
             self._num_vin = compact.read_from(self.stream)
         return self._num_vin
 
@@ -80,7 +77,7 @@ class GlobalTransactionView:
     @property
     def vin0_offset(self):
         if self._vin0_offset is None:
-            self._vin0_offset = self.offset + 4 + len(compact.to_bytes(self.num_vin))
+            self._vin0_offset = self.offset + self.NUM_VIN_OFFSET + len(compact.to_bytes(self.num_vin))
         return self._vin0_offset
 
     @property
@@ -154,9 +151,15 @@ class PSBTView:
         self.compress = compress
         self._tx_version = self.tx.version if self.tx else None
         self._locktime = self.tx.locktime if self.tx else None
+        self.clear_cache()
+
+    def clear_cache(self):
+        # cache for digests
         self._hash_prevouts = None
-        self._hash_outputs = None
         self._hash_sequence = None
+        self._hash_outputs = None
+        self._hash_amounts = None
+        self._hash_script_pubkeys = None
 
     @classmethod
     def view(cls, stream, offset=None, compress=False):
@@ -212,13 +215,16 @@ class PSBTView:
                    first_scope, version, tx_offset, compress)
 
     def _skip_scope(self):
+        off = 0
         while True:
             # read key and update cursor
-            key = read_string(self.stream)
-            # separator
-            if len(key) == 0:
+            keylen = skip_string(self.stream)
+            off += keylen
+            # separator: zero-length key
+            if keylen == 1:
                 break
-            skip_string(self.stream)
+            off += skip_string(self.stream)
+        return off
 
     def seek_to_scope(self, n):
         """
@@ -226,17 +232,22 @@ class PSBTView:
         n can be from 0 to num_inputs+num_outputs or None.
         If n = None it seeks to global scope.
         If n = num_inputs + num_outputs it seeks to the end of PSBT.
-        This can be useful to check that nothing is left in the stream (i.e. for tests)
+        This can be useful to check that nothing is left in the stream (i.e. for tests).
+        Returns an offset at this scope.
         """
         if n is None:
-            self.stream.seek(self.offset)
+            off = self.offset+len(self.MAGIC)
+            self.stream.seek(off)
+            return off
         if n > self.num_inputs + self.num_outputs:
             raise PSBTError("Invalid scope number")
         # seek to first scope
         self.stream.seek(self.first_scope)
+        off = self.first_scope
         while n:
-            self._skip_scope()
+            off += self._skip_scope()
             n -= 1
+        return off
 
     def input(self, i):
         """Reads, parses and returns PSBT InputScope #i"""
@@ -261,15 +272,15 @@ class PSBTView:
             return self.tx.vin(i)
 
         self.seek_to_scope(i)
-        v = self._get_value(b"\x0e", from_current=True)
+        v = self.get_value(b"\x0e", from_current=True)
         txid = bytes(reversed(v))
 
         self.seek_to_scope(i)
-        v = self._get_value(b"\x0f", from_current=True)
+        v = self.get_value(b"\x0f", from_current=True)
         vout = int.from_bytes(v, 'little')
 
         self.seek_to_scope(i)
-        v = self._get_value(b"\x10", from_current=True) or b"\xFF\xFF\xFF\xFF"
+        v = self.get_value(b"\x10", from_current=True) or b"\xFF\xFF\xFF\xFF"
         sequence = int.from_bytes(v, 'little')
 
         return TransactionInput(txid, vout, sequence=sequence)
@@ -281,11 +292,11 @@ class PSBTView:
             return self.tx.vout(i)
 
         self.seek_to_scope(self.num_inputs + i)
-        v = self._get_value(b"\x03", from_current=True)
+        v = self.get_value(b"\x03", from_current=True)
         value = int.from_bytes(v, 'little')
 
         self.seek_to_scope(self.num_inputs + i)
-        v = self._get_value(b"\x04", from_current=True)
+        v = self.get_value(b"\x04", from_current=True)
         script_pubkey = Script(v)
 
         return TransactionOutput(value, script_pubkey)
@@ -293,31 +304,44 @@ class PSBTView:
     @property
     def locktime(self):
         if self._locktime is None:
-            v = self._get_value(b"\x03")
+            v = self.get_value(b"\x03")
             self._locktime = int.from_bytes(v, 'little') if v is not None else 0
         return self._locktime
 
     @property
     def tx_version(self):
         if self._tx_version is None:
-            v = self._get_value(b"\x02")
+            v = self.get_value(b"\x02")
             self._tx_version = int.from_bytes(v, 'little') if v is not None else 0
         return self._tx_version
 
-    def _get_value(self, key_start, from_current=False):
+    def seek_to_value(self, key_start, from_current=False):
+        """
+        Seeks to value with key starting with key_start.
+        Returns offset - relative if from_current=True, absolute otherwise.
+        If key is not found - returns None.
+        """
+        off = 0
         if not from_current:
             # go to the start
             self.stream.seek(self.offset + len(self.MAGIC))
+            off = self.offset + len(self.MAGIC)
         while True:
             key = read_string(self.stream)
+            off += len(key) + len(compact.to_bytes(len(key)))
             # separator - not found
             if len(key) == 0:
                 return None
             # matches
             if key.startswith(key_start):
-                return read_string(self.stream)
+                return off
             # continue to the next key
-            skip_string(self.stream)
+            off += skip_string(self.stream)
+
+    def get_value(self, key_start, from_current=False):
+        off = self.seek_to_value(key_start, from_current)
+        if off:
+            return read_string(self.stream)
 
     def hash_prevouts(self):
         if self._hash_prevouts is None:
@@ -347,11 +371,57 @@ class PSBTView:
             self._hash_outputs = h.digest()
         return self._hash_outputs
 
+    def hash_amounts(self, amounts):
+        if self._hash_amounts is None:
+            self._hash_amounts = hash_amounts(amounts)
+        return self._hash_amounts
+
+    def hash_script_pubkeys(self, script_pubkeys):
+        if self._hash_script_pubkeys is None:
+            self._hash_script_pubkeys = hash_script_pubkeys(script_pubkeys)
+        return self._hash_script_pubkeys
+
+    def sighash_taproot(self, input_index, script_pubkeys, values, sighash=SIGHASH.DEFAULT):
+        """check out bip-341"""
+        if input_index < 0 or input_index >= self.num_inputs:
+            raise PSBTError("Invalid input index")
+        if len(values) != self.num_inputs:
+            raise PSBTError("All spent amounts are required")
+        sh, anyonecanpay = SIGHASH.check(sighash)
+        h = hashes.tagged_hash_init("TapSighash", b"\x00")
+        h.update(bytes([sighash]))
+        h.update(self.tx_version.to_bytes(4, "little"))
+        h.update(self.locktime.to_bytes(4, "little"))
+        if not anyonecanpay:
+            h.update(self.hash_prevouts())
+            h.update(self.hash_amounts(values))
+            h.update(self.hash_script_pubkeys(script_pubkeys))
+            h.update(self.hash_sequence())
+        if sh not in [SIGHASH.SINGLE, SIGHASH.NONE]:
+            h.update(self.hash_outputs())
+        # data about this input
+        h.update(b"\x00") # ext_flags and annex are not supported
+        if anyonecanpay:
+            vin = self.vin(input_index)
+            h.update(vin.serialize())
+            h.update(values[input_index].to_bytes(8, "little"))
+            h.update(script_pubkeys[input_index].serialize())
+            h.update(vin.sequence.to_bytes(4, "little"))
+        else:
+            h.update(input_index.to_bytes(4, "little"))
+        # annex is not supported
+        if sh == SIGHASH.SINGLE:
+            h.update(self.vout(input_index).serialize())
+        return h.digest()
+
     def sighash_segwit(self, input_index, script_pubkey, value, sighash=SIGHASH.ALL):
         """check out bip-143"""
         if input_index < 0 or input_index >= self.num_inputs:
             raise PSBTError("Invalid input index")
         sh, anyonecanpay = SIGHASH.check(sighash)
+        # default sighash is used only in taproot
+        if sh == SIGHASH.DEFAULT:
+            sh = SIGHASH.ALL
         inp = self.vin(input_index)
         zero = b"\x00"*32 # for sighashes
         h = hashlib.sha256()
@@ -371,7 +441,7 @@ class PSBTView:
         h.update(inp.sequence.to_bytes(4, "little"))
         if not (sh in [SIGHASH.NONE, SIGHASH.SINGLE]):
             h.update(hashlib.sha256(self.hash_outputs()).digest())
-        elif sh == SIGHASH.SINGLE and input_index < len(self.vout):
+        elif sh == SIGHASH.SINGLE and input_index < self.num_outputs:
             h.update(hashlib.sha256(
                 hashlib.sha256(self.vout(input_index).serialize()).digest()
             ).digest())
@@ -385,6 +455,9 @@ class PSBTView:
         if input_index < 0 or input_index >= self.num_inputs:
             raise PSBTError("Invalid input index")
         sh, anyonecanpay = SIGHASH.check(sighash)
+        # default sighash is used only in taproot
+        if sh == SIGHASH.DEFAULT:
+            sh = SIGHASH.ALL
         # no corresponding output for this input, we sign 00...01
         if sh == SIGHASH.SINGLE and input_index >= self.num_outputs:
             return b"\x00"*31+b"\x01"
@@ -430,6 +503,16 @@ class PSBTView:
     def sighash(self, i, sighash=SIGHASH.ALL, input_scope=None):
         inp = self.input(i) if input_scope is None else input_scope
 
+        if inp.is_taproot:
+            # TODO: not very optimal. Maybe use build_cache() or something?
+            values = []
+            scripts = []
+            for idx in range(self.num_inputs):
+                inp = self.input(idx)
+                values.append(inp.utxo.value)
+                scripts.append(inp.utxo.script_pubkey)
+            return self.sighash_taproot(i, script_pubkeys=scripts, values=values, sighash=sighash)
+
         value = inp.utxo.value
         sc = inp.witness_script or inp.redeem_script or inp.utxo.script_pubkey
 
@@ -452,7 +535,7 @@ class PSBTView:
             h = self.sighash_legacy(i, sc, sighash=sighash)
         return h
 
-    def sign_input(self, i, root, sig_stream, sighash=SIGHASH.ALL, extra_scope_data=None) -> int:
+    def sign_input(self, i, root, sig_stream, sighash=SIGHASH.DEFAULT, extra_scope_data=None) -> int:
         """
         Signs input taking into account additional derivation information for this input.
         It's helpful if your wallet knows more than provided in PSBT.
@@ -471,10 +554,19 @@ class PSBTView:
         inp = self.input(i)
         if extra_scope_data is not None:
             inp.update(extra_scope_data)
+
+        # SIGHASH.DEFAULT is only for taproot, fallback to SIGHASH.ALL for other inputs
+        required_sighash = sighash
+        if not inp.is_taproot and required_sighash == SIGHASH.DEFAULT:
+            required_sighash = SIGHASH.ALL
+
         # check which sighash to use
-        inp_sighash = inp.sighash_type or sighash or SIGHASH.ALL
-        # if input sighash is set and is different from kwarg - skip input
-        if sighash is not None and inp_sighash != sighash:
+        inp_sighash = inp.sighash_type or required_sighash or SIGHASH.DEFAULT
+        if not inp.is_taproot and inp_sighash == SIGHASH.DEFAULT:
+            inp_sighash = SIGHASH.ALL
+
+        # if input sighash is set and is different from kwarg - don't sign this input
+        if required_sighash is not None and inp_sighash != required_sighash:
             return 0
 
         h = self.sighash(i, sighash=inp_sighash, input_scope=inp)
@@ -483,6 +575,45 @@ class PSBTView:
 
         counter = 0
         partial_sigs = OrderedDict()
+
+        # taproot is special
+        # currently works only for single key
+        if inp.is_taproot:
+            # individual private key
+            if not fingerprint:
+                # TODO: tweak using taproot psbt fields
+                pk = root.taproot_tweak(b"")
+                if pk.xonly() in sc.data:
+                    sig = pk.schnorr_sign(h)
+                    wit = sig.serialize()
+                    if inp_sighash != SIGHASH.DEFAULT:
+                        wit += bytes([inp_sighash])
+                    inp.final_scriptwitness = Witness([wit])
+                    counter = 1
+            # if we use HDKey
+            else:
+                # TODO: add taproot derivation paths and scripts
+                for pub in inp.bip32_derivations:
+                    # check if it is root key
+                    if inp.bip32_derivations[pub].fingerprint == fingerprint:
+                        hdkey = root.derive(inp.bip32_derivations[pub].derivation)
+                        mypub = hdkey.key.get_public_key()
+                        if mypub != pub:
+                            raise PSBTError("Derivation path doesn't look right")
+                        pk = hdkey.taproot_tweak(b"")
+                        if pk.xonly() in sc.data:
+                            sig = pk.schnorr_sign(h)
+                            # sig plus sighash flag
+                            wit = sig.serialize()
+                            if inp_sighash != SIGHASH.DEFAULT:
+                                wit += bytes([inp_sighash])
+                            inp.final_scriptwitness = Witness([wit])
+                            counter = 1
+            if counter:
+                ser_string(sig_stream, b"\x08")
+                ser_string(sig_stream, inp.final_scriptwitness.serialize())
+            return counter
+
         # if we have individual private key
         if not fingerprint:
             # check if we are included in the script
@@ -509,12 +640,12 @@ class PSBTView:
             ser_string(sig_stream, partial_sigs[pub])
         return counter
 
-    def sign_with(self, root, sig_stream, sighash=SIGHASH.ALL) -> int:
+    def sign_with(self, root, sig_stream, sighash=SIGHASH.DEFAULT) -> int:
         """
         Signs psbtview with root key (HDKey or similar) and writes per-input signatures to the sig_stream.
         It can be either a simple BytesIO object or a file stream open for writing.
         Returns number of signatures added to PSBT.
-        Sighash kwarg is set to SIGHASH.ALL by default,
+        Sighash kwarg is set to SIGHASH.DEFAULT, for segwit and legacy it's replaced to SIGHASH.ALL
         so if PSBT is asking to sign with a different sighash this function won't sign.
         If you want to sign with sighashes provided in the PSBT - set sighash=None.
         """
